//------------------------------------------------------------------------
//  ABOUT WINDOW
//------------------------------------------------------------------------
//
//  Eureka DOOM Editor
//
//  Copyright (C) 2001-2016 Andrew Apted
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//------------------------------------------------------------------------

#ifndef __EUREKA_UI_ABOUT_H__
#define __EUREKA_UI_ABOUT_H__

void InitAboutDialog();

void DLG_AboutText(void);

#endif /* __EUREKA_UI_ABOUT_H__ */

//--- editor settings ---
// vi:ts=4:sw=4:noexpandtab
