/*** this CSS is based on pmwiki.css ***/

/* This sets the overall frame for the site */
body {
  margin:0px; background-color: black;
  font-family:Arial,Helvetica,sans-serif; font-size:14pt;
  color: #aaa;
}

/* These control the fixed-width text elements of the page */
textarea, pre, code { font-size:0.9em; }
pre, code { font-family:'Lucida Console','Andale Mono','Courier New',Courier,monospace; }
pre { line-height:1.2em; }
pre code, code code, pre pre { font-size:100%; }

/* These primarily adjust the size and spacing of heading elements,
** most browsers have atrocious defaults for these. */
h1, h2, h3, h4, h5, h6 { margin-top:1.0em; margin-bottom:0.6em; }
h1, h2, h3, h6 { font-weight:normal; }
h4, h5 { font-weight:bold; }
h1 code, h2 code, h3 code, h4 code { font-size:1em; }
h1 { font-size:1.90em; }
h2 { font-size:1.50em; }
h3 { font-size:1.25em; }
h4 { font-size:1.10em; }
h5 { font-size:1.0em; }
h6 { font-size:1.0em; }

#wikimid {
  table-layout: fixed;
}

/* The #wikilogo element is the logo from $PageLogoFmt */
#wikilogo {
  width=240px; margin:0px; padding:0px;
  background-color: black;
  border-bottom: 1px #aaa solid;
  text-align: center;
}

/* This controls the rest of the heading (primarily the search box) */
#wikihead {
  margin: 0px;
/*  position:absolute; right:10px; top:10px;
  font-family:Verdana,sans-serif; font-size:85%;  */
}

/* These are for the left-sidebar. */
#wikileft {
  width:     250px;
  min-width: 250px;
  max-width: 250px;
  margin: 0px;
  overflow: hidden;
  background-image: url("http://eureka-editor.sourceforge.net/grid.png");
  background-position: top;
  background-color: #002;
}

#wikileft2 {
  padding:5px; border-right:1px #aaa solid;
  padding-top: 15px;
  line-height:1.33em;
  font-size:14pt; font-family:Verdana,sans-serif;
}
#wikileft2 .vspace { margin-top:1.125em; }
#wikileft2 a { text-decoration:none; color:#09c; }
#wikileft2 a:hover { text-decoration:underline; color:#ee0; }
#wikileft2 ul { list-style:none; padding:0px; margin:0px; }
#wikileft2 li { margin:0px; padding-left: 6px; }
.sidehead {
  margin:0px; padding:4px 2px 2px 2px;
  font-size:14pt; font-weight:bold; font-style:normal;
}
.sidehead a
  { color:#505050; font-weight:bold; font-style:normal; }

/* These affect the main content area. */
#wikibody {
  margin: 0px;
  padding: 10px 10px 20px 20px;
  background-color: #f5f5f5;
  color: black;
  font-size:14pt;
}

#wikibody a:link    { color:blue; text-decoration:none; }
#wikibody a:visited { color:blue; text-decoration:none; }
#wikibody a:hover   { color:#07f; text-decoration:underline; }

#wikicmds {
  float:right; white-space:nowrap;
  font-family:Verdana,sans-serif; font-size:80%;
}

#wikicmds a:link    { color:#666; text-decoration:none; border:none; }
#wikicmds a:visited { color:#666; text-decoration:none; border:none; }
#wikicmds a:hover { text-decoration:underline; color:blue; }
#wikicmds a.createlink { display:none; }

#wikicmds ul { list-style:none; margin:0px; padding:0px; }
#wikicmds li { display:inline; margin:0px 5px; }
.pagegroup { margin-top:8px; margin-bottom:2px; }
.pagetitle { line-height:1em; margin:0px; font-size:1.6em; font-weight:normal; }
.wikiaction { margin-top:4px; margin-bottom:4px; }
#wikitext { margin-top:12px; line-height:1.33em; }
#wikitext table { font-size:100%; line-height:1.33em; } /* For MSIE 5.5 */

/* These are for the edit form. */
#wikiedit form { margin:0px; width:100%; }
#wikiedit textarea { width:100%; }
.wikimessage { margin-top:4px; margin-bottom:4px; font-style:italic; }

/* These affect the lines at the very bottom. */
#wikifoot {
  padding-left:256px;
  padding-bottom:4px; border-top:1px #aaa solid;
  padding-top:0px;
  font-family:Verdana,sans-serif; font-size:80%;
}
#wikifoot input { font-size:85%; }
#wikifoot a { color:#eb4; }

/* andrewj's stuff */
.footerlinks { float: right; }

.footnav { padding-top:4px; padding-bottom:10px }

#cheatsheet {
    font-size: 85%;
    background-color: #ececfc;
    border: 1px #7e7e6e solid;
    padding: 0.3em;
    margin: 0.9em;
    line-height: 1.33em;
}

.cheatsheet ul { list-style:none; }

.cheatbold { color: black; font-weight: bold; }

.key {
    background-color: white;
    font-family: monospace;
    font-size: 90%;
    border: 1px #aaa solid;
    padding: 3px 6px 3px 6px;
}

#codebox {
    background-color: #fcfcec;
    font-family: monospace;
    font-size: 80%;
    border: 1px #8e8e7e solid;
    padding: 0.3em;
    margin: 0.6em;
    margin-left: 3.5em;
}

#noticebox {
    background-color: #ececfc;
    border: 1px #7e7e8e solid;
    padding: 0.3em;
    margin: 0.6em;
}

#warningbox {
    background-color: #fcdcdc;
    border: 1px #aa9999 solid;
    padding: 0.3em;
    margin: 0.6em;
}

.kw { color: #d00; font-family: monospace; }

/* These affect the printed appearance of the web view (not the separate
** print view) of pages.  The sidebar and action links aren't printed. */
@media print {
  body { width:auto; margin:0px; padding:0.5em; }
  #wikihead, #wikilogo, #wikileft2, #wikicmds, .footnav { display:none; }
  #wikifoot { padding:2px; }
}

